/* MoarNES (c)2010-2011 Mike Chambers
   http://moarnes.rubbermallet.org

   A Nintendo Entertainment System emulator.

   This open-source software is (c)2010-2011 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.
*/

#include <stdio.h>
#include "config.h"
#include "common.h"
#include "stdint.h"
#include "input.h"

const uint8_t *build = "v0.11.12.12 alpha";
bool ROMwaiting = false;
bool exitapp = false;
uint8_t nextROM[512];
uint8_t app_path[MAX_PATH];

#ifdef _WIN32
	HANDLE GameThread;
	HANDLE NESThread;
#endif
extern bool initSDL(uint8_t *windowcaption);
extern uint8_t *readSDLerr();
extern DWORD WINAPI threadNES(LPVOID lParam);
extern void execNES();
extern bool loadROM(uint8_t *filename);

extern void initPPU();
extern void killPPU();
extern void initinput();
extern void blitframe(uint8_t *NESbuffer);

#ifndef OLD_6502
	extern void hookexternal(void *funcptr);
#else
	extern void init6502();
#endif
extern void processinput();
extern void reset6502();
extern bool initAPU(uint8_t CPUmode, uint32_t samplerate);
extern void tickchannelsAPU();
extern void drawtext(uint8_t *text, uint16_t, uint16_t y);

extern uint8_t dorecord;
extern uint8_t recordpath[256];
extern uint8_t outputNES[240][256];

extern bool running;
extern bool paused;
extern uint32_t cursamplerate;

extern int16_t msgbox(uint8_t *msgtext, uint16_t buttons);

void loadconfig();
void saveconfig();
int main(int argc, char *argv[]) {
	uint8_t textdata[256] = { 0 };
	int16_t charfind;

	charfind = (int16_t)GetModuleFileName(NULL, &app_path[0], MAX_PATH);
	for (; charfind>=0; charfind--) {
		if (app_path[charfind] == '\\') {
			app_path[charfind+1] = 0;
			break;
		}
	}

	sprintf(&textdata[0], "MoarNES %s", build);
	if (initSDL(&textdata[0]) == false) return(-1);
	if (initAPU(1, 96000) == false) if (initAPU(1, 48000) == false) if (initAPU(1, 44100) == false)
		if (initAPU(1, 32000) == false) if (initAPU(1, 22050) == false) if (initAPU(1, 11025) == false) return(-1);
	initPPU();

	#ifndef OLD_6502
		hookexternal(&tickchannelsAPU);
	#else
		init6502();
	#endif
	initinput();
	loadconfig();

	memset(&outputNES[0][0], 65, 61440); //blank screen to black
	drawtext(&textdata[0], (uint16_t)(128-(strlen(&textdata[0])*8)/2), 224);
	sprintf(&textdata[0], "Audio initialized at %lu Hz", cursamplerate);
	drawtext(&textdata[0], (uint16_t)(128-(strlen(&textdata[0])*8)/2), 232);
	blitframe(&outputNES[0][0]);

	while(exitapp == false) {
		processinput();
		if (running == true) execNES();
		if (ROMwaiting == true) {
			loadROM(&nextROM[0]);
			running = true;
		}
		if (running == false) processinput();
		SDL_Delay(10);
	}

	saveconfig();
	return(0);
}

#ifdef _WIN32
extern HWND myWindow;
int16_t msgbox(uint8_t *msgtxt, uint16_t buttons) {
	return(MessageBox(myWindow, msgtxt, "MoarNES", buttons));
}

void fatalerr(uint8_t *errmsg) {
	uint8_t newmsg[1024];

	sprintf(&newmsg[0], "FATAL ERROR: %s", errmsg);
	MessageBox(myWindow, &newmsg[0], "Fatal error", MB_ICONERROR | MB_OK);
	exit(-1);
}
#endif

bool verifyexit() {
	switch (msgbox("Are you sure you want to exit MoarNES?", MB_ICONQUESTION | MB_YESNO)) {
		case IDYES:
			running = false;
			paused = false;
			exitapp = true;
			return(true);
	}

	return(false);
}

extern uint8_t app_path[MAX_PATH];
extern struct inputdev_s gamepad[2];
extern uint16_t screenmode;
extern void setscreenmode(uint16_t modeval);
#ifdef _WIN32
extern HWND myWindow;
RECT winrect;
#else
uint8_t blank[4] = { 0, 0, 0, 0 };
#endif
void loadconfig() {
	FILE *configfile = NULL;
	uint8_t path[MAX_PATH];
	int16_t wx, wy;

	sprintf(&path[0], "%sMoarNES.conf", &app_path[0]);
	configfile = fopen(&path[0], "rb");
	if (configfile == NULL) {
		msgbox("This appears to be the first time you have run MoarNES, or the configuration file has been deleted. A new one will be created in the application's folder using the default options.", MB_ICONINFORMATION | MB_OK);
		saveconfig();
		return;
	}

	fread(&screenmode, 1, 2, configfile);
	fread(&wx, 1, 2, configfile);
	fread(&wy, 1, 2, configfile);
	#ifdef _WIN32
		SetWindowPos(myWindow, NULL, wx, wy, 0, 0, SWP_NOSIZE);
	#endif
	fread(&gamepad[0], 1, sizeof(struct inputdev_s) << 1, configfile);
	fclose(configfile);

	setscreenmode(screenmode);
}

void saveconfig() {
	FILE *configfile = NULL;
	uint8_t path[MAX_PATH];

	sprintf(&path[0], "%sMoarNES.conf", &app_path[0]);
	configfile = fopen(&path[0], "wb");
	if (configfile == NULL) return;

	fwrite(&screenmode, 1, 2, configfile);
	#ifdef _WIN32
		GetWindowRect(myWindow, (LPRECT)&winrect);
		fwrite(&winrect.left, 1, 2, configfile);
		fwrite(&winrect.top, 1, 2, configfile);
	#else
		fwrite(&blank[0], 1, 4, configfile);
	#endif
	fwrite(&gamepad[0], 1, sizeof(struct inputdev_s) << 1, configfile);

	fclose(configfile);
}
