#ifdef _WIN32
#include <stdio.h>
#include "common.h"
#include "stdint.h"
#include "rom.h"

uint8_t msgdata[1024];

//extern uint8_t dodump;
extern uint8_t romloaded;
extern uint8_t rompath[MAX_PATH];
extern SDL_Surface *screen;
extern const uint8_t *build;
extern int16_t msgbox(uint8_t *msgtxt, uint16_t buttons);
extern void fatalerr(uint8_t *errmsg);
extern void reset6502();
extern bool loadROM(uint8_t *filename);
extern DWORD WINAPI threadNES(LPVOID lParam);
extern void verifyexit();
extern void execNES();
extern bool stopNES();
extern bool initAPU(uint8_t CPUmode, uint32_t samplerate);
extern void killAPU();
extern bool running;
extern bool ROMwaiting;
extern bool paused;
extern bool exitapp;
extern uint32_t cursamplerate;
extern uint8_t nextROM[512];
extern void setscreenmode(uint16_t modeval);
extern void configurekeyboard(uint8_t controller);
extern void configurejoystick(uint8_t controller);
extern uint8_t allowsweeps;
extern uint8_t savestatefilename[MAX_PATH];
extern bool doloadstate;
extern bool dosavestate;
extern uint8_t gamegenie;
extern uint8_t joycount;
extern struct cart_s cartridge;
extern void timedtext(uint8_t *text);
extern void drawtext(uint8_t *text, uint16_t x, uint16_t y);
extern void blitframe(uint8_t *NESbuffer);
extern uint8_t outputNES[240][256];

void showabout();

void MenuItemClick(WPARAM wParam) {
	OPENFILENAME of_dlg;
	uint8_t filename[MAX_PATH] = { 0 };
	uint8_t textdata[256];

	switch (LOWORD(wParam)) {
		//file menu
		case ID_FILE_OPENROM:
		case ID_FILE_OPENROMWITHGAMEGENIE:
			memset(&of_dlg, 0, sizeof(of_dlg));
			of_dlg.lStructSize = sizeof(of_dlg);
			of_dlg.lpstrTitle = "Open ROM";
			of_dlg.hInstance = NULL;
			of_dlg.lpstrFile = filename;
			of_dlg.lpstrFilter = "NES ROM images (*.nes)\0*.nes\0\0";
			of_dlg.nMaxFile = MAX_PATH;
			of_dlg.Flags = OFN_FILEMUSTEXIST | OFN_LONGNAMES;
			if (GetOpenFileName(&of_dlg)) {
				running = false;
				if (LOWORD(wParam) == ID_FILE_OPENROMWITHGAMEGENIE) gamegenie = 1;
					else gamegenie = 0;
				if (loadROM(of_dlg.lpstrFile) == false) return;
				sprintf(&nextROM[0], of_dlg.lpstrFile);
				ROMwaiting = true;
				timedtext("ROM loaded.");
			}
			break;

		case ID_FILE_CLOSEROM:
			running = false;
			ROMwaiting = false;
			break;

		case ID_FILE_ROMINFORMATION:
			if (!romloaded) {
				msgbox("No ROM is currently loaded.", MB_ICONINFORMATION | MB_OK);
			} else {
				sprintf(&textdata[0], "ROM information:\n\nFile: %s\nMapper: %u\nPRG data: %u KB\nCHR data: %u KB", &rompath[0], cartridge.mapper, cartridge.PRGcount, cartridge.CHRcount);
				msgbox(&textdata[0], MB_ICONINFORMATION | MB_OK);
			}
			break;
		
		case ID_FILE_LOADSTATE:
			memset(&of_dlg, 0, sizeof(of_dlg));
			of_dlg.lStructSize = sizeof(of_dlg);
			of_dlg.lpstrTitle = "Open state";
			of_dlg.hInstance = NULL;
			of_dlg.lpstrFile = filename;
			of_dlg.lpstrFilter = "MoarNES save states (*.mns)\0*.mns\0\0";
			of_dlg.nMaxFile = MAX_PATH;
			of_dlg.Flags = OFN_FILEMUSTEXIST | OFN_LONGNAMES;
			if (GetOpenFileName(&of_dlg)) {
				sprintf(&savestatefilename[0], of_dlg.lpstrFile);
				doloadstate = true;
			}
			break;

		case ID_FILE_SAVESTATE:
			memset(&of_dlg, 0, sizeof(of_dlg));
			of_dlg.lStructSize = sizeof(of_dlg);
			of_dlg.lpstrTitle = "Save state";
			of_dlg.hInstance = NULL;
			of_dlg.lpstrFile = filename;
			of_dlg.lpstrFilter = "MoarNES save states (*.mns)\0*.mns\0\0";
			of_dlg.nMaxFile = MAX_PATH;
			of_dlg.Flags = OFN_LONGNAMES;
			if (GetSaveFileName(&of_dlg)) {
				sprintf(&savestatefilename[0], "%s.mns", of_dlg.lpstrFile);
				if (memcmp(&savestatefilename[strlen(&savestatefilename[0]) - 8], ".mns.mns", 4) == 0) {
					savestatefilename[strlen(&savestatefilename[0]) - 4] = 0;
				}
				dosavestate = true;
			}
			break;

		case ID_FILE_EXIT:
			verifyexit();
			break;

		//emulation menu
		case ID_CPU_RUN:
			if (paused == true) timedtext("CPU resumed.");
			paused = false;
			break;

		case ID_CPU_PAUSE:
			if (paused == false) {
				timedtext("CPU paused.");
				drawtext("CPU paused.", 0, 232);
				blitframe(&outputNES[0][0]);
			}
			paused = true;
			break;

		case ID_CPU_RESET:
			reset6502();
			break;

		case ID_GAMEPAD1_CONFIGUREKEYBOARD:
			configurekeyboard(0);
			break;

		case ID_GAMEPAD1_CONFIGUREJOYSTICK:
			if (joycount) configurejoystick(0);
				else msgbox("There are no joysticks connected to this computer!", MB_ICONINFORMATION | MB_OK);
			break;

		case ID_GAMEPAD2_CONFIGUREKEYBOARD:
			configurekeyboard(1);
			break;

		case ID_GAMEPAD2_CONFIGUREJOYSTICK:
			if (joycount) configurejoystick(1);
				else msgbox("There are no joysticks connected to this computer!", MB_ICONINFORMATION | MB_OK);
			break;

		//video menu
		case ID_ZOOM_1X:
			setscreenmode(1);
			break;

		case ID_ZOOM_2X:
			setscreenmode(2);
			break;

		case ID_ZOOM_3X:
			setscreenmode(3);
			break;

		case ID_ZOOM_4X:
			setscreenmode(4);
			break;

		case ID_ZOOM_FULLSCREEN:
			setscreenmode(5);
			break;

		//audio menu
		case ID_SAMPLERATE_96000HZ:
			killAPU();
			if (initAPU(1, 96000) == false) {
				sprintf(&msgdata[0], "Unable to set the selected samplerate!\nStill using old sample rate of %lu Hz.\n", cursamplerate);
				msgbox(&msgdata[0], MB_ICONWARNING | MB_OK);
				if (initAPU(1, cursamplerate) == false) {
					fatalerr("Double failure during audio initialization!");
				}
			}
			break;

		case ID_SAMPLERATE_48000HZ:
			killAPU();
			if (initAPU(1, 48000) == false) {
				sprintf(&msgdata[0], "Unable to set the selected samplerate!\nStill using old sample rate of %lu Hz.\n", cursamplerate);
				msgbox(&msgdata[0], MB_ICONWARNING | MB_OK);
				if (initAPU(1, cursamplerate) == false) {
					fatalerr("Double failure during audio initialization!");
				}
			}
			break;

		case ID_SAMPLERATE_44100HZ:
			killAPU();
			if (initAPU(1, 44100) == false) {
				sprintf(&msgdata[0], "Unable to set the selected samplerate!\nStill using old sample rate of %lu Hz.\n", cursamplerate);
				msgbox(&msgdata[0], MB_ICONWARNING | MB_OK);
				if (initAPU(1, cursamplerate) == false) {
					fatalerr("Double failure during audio initialization!");
				}
			}
			break;

		case ID_SAMPLERATE_32000HZ:
			killAPU();
			if (initAPU(1, 32000) == false) {
				sprintf(&msgdata[0], "Unable to set the selected samplerate!\nStill using old sample rate of %lu Hz.\n", cursamplerate);
				msgbox(&msgdata[0], MB_ICONWARNING | MB_OK);
				if (initAPU(1, cursamplerate) == false) {
					fatalerr("Double failure during audio initialization!");
				}
			}
			break;

		case ID_SAMPLERATE_22050HZ:
			killAPU();
			if (initAPU(1, 22050) == false) {
				sprintf(&msgdata[0], "Unable to set the selected samplerate!\nStill using old sample rate of %lu Hz.\n", cursamplerate);
				msgbox(&msgdata[0], MB_ICONWARNING | MB_OK);
				if (initAPU(1, cursamplerate) == false) {
					fatalerr("Double failure during audio initialization!");
				}
			}
			break;

		case ID_SAMPLERATE_11025HZ:
			killAPU();
			if (initAPU(1, 11025) == false) {
				sprintf(&msgdata[0], "Unable to set the selected samplerate!\nStill using old sample rate of %lu Hz.\n", cursamplerate);
				msgbox(&msgdata[0], MB_ICONWARNING | MB_OK);
				if (initAPU(1, cursamplerate) == false) {
					fatalerr("Double failure during audio initialization!");
				}
			}
			break;

		/*case ID_AUDIO_ENABLESWEEPUNITS:
			allowsweeps = 1;
			break;

		case ID_AUDIO_DISABLESWEEPUNITS:
			allowsweeps = 0;
			break;*/

        //help menu
		case ID_HELP_ABOUT:
			showabout();
			break;
	}
}

void showabout() {
	uint8_t about[512];

	sprintf(&about[0], "MoarNES %s (c)2010-2011 Mike Chambers\nMoarNES is distributed under the terms of the GNU GPL v2.\n\nHome page: http://moarnes.sourceforge.net", build);
	msgbox(&about[0], MB_ICONINFORMATION | MB_OK);
}
#endif

