/* MoarNES source - old6502.c

   This open-source software is (c)2010-2011 Mike Chambers, and is released
   under the terms of the GNU GPL v2 license.

   This file contains 6502 emulator code by Neil Bradley. It's enabled if OLD_6502
   is defined in config.h. I am going to be removing it soon. Just use the 6502 emu
   I wrote myself in 6502.c. The timing in this one here is not very accurate.
*/

#include "config.h"

#ifdef OLD_6502
#include <stdio.h>
#include "stdint.h"

#define WORD unsigned short
#define BYTE unsigned char

extern uint8_t read6502(uint16_t addr);
extern void write6502(uint16_t addr, uint8_t value);

/* flags = NVRBDIZC */
BYTE a_reg,x_reg,y_reg,flag_reg,s_reg;
WORD pc_reg = 0;

/* Macros for convenience */
#define A a_reg
#define X x_reg
#define Y y_reg
#define P flag_reg
#define S s_reg
#define PC pc_reg

/* internal registers */
BYTE opcode;
uint32_t clockticks6502 = 0;
uint32_t totalticks = 0;
extern long frameticks;
extern long scanlineticks;
long remainTicks = 0;
BYTE curkey;
WORD tmpspr;
extern char doaudio;
extern WORD freqmod;

/* help variables */
WORD savepc;
BYTE value;
int sum,saveflags;

/* arrays */
void (*adrmode[256])();
void (*instruction[256])();
int ticks[256];
WORD help;

/* Adressing modes */
/*#define implied6502() { break; }
#define immediate6502() { savepc=PC++; }
#define abs6502() { savepc = read6502(PC) + (read6502(PC + 1) << 8); PC++; PC++; }
#define relative6502() { savepc = read6502(PC++); if (savepc & 0x80) savepc -= 0x100; if ((savepc>>8) != (PC>>8)) clockticks6502++; }
#define indirect6502() { help = read6502(PC) + ((read6502(PC + 1)&0xFF) << 8); if ((help&0xFF)==0xFF) savepc = read6502(help) + (read6502(help&0xFF00) << 8); else savepc = read6502(help) + (read6502(help + 1) << 8); PC++; PC++; }
#define absx6502() { savepc = read6502(PC) + (read6502(PC + 1) << 8); PC++; PC++; if (ticks[opcode]==4) if ((savepc>>8) != ((savepc+X)>>8)) clockticks6502++; savepc += X; }
#define absy6502() { savepc = read6502(PC) + (read6502(PC + 1) << 8); PC++; PC++; if (ticks[opcode]==4) if ((savepc>>8) != ((savepc+Y)>>8)) clockticks6502++; savepc += Y; }
#define zp6502() { savepc=read6502(PC++); }
#define zpx6502() { savepc=read6502(PC++)+X; savepc &= 0x00ff; }
#define zpy6502() { savepc=read6502(PC++)+Y; savepc &= 0x00ff; }
#define indx6502() { value = read6502(PC++)+X; savepc = read6502(value) + (read6502(value+1) << 8); }
#define indy6502() { value = read6502(PC++); savepc = read6502(value) + (read6502(value+1) << 8); if (ticks[opcode]==5) if ((savepc>>8) != ((savepc+Y)>>8)) clockticks6502++; savepc += Y; }
#define indabsx6502() { help = read6502(PC) + (read6502(PC + 1) << 8) + X; savepc = read6502(help) + (read6502(help + 1) << 8); }
#define indzp6502() { value = read6502(PC++); savepc = read6502(value) + (read6502(value + 1) << 8); }*/

void implied6502()
{
}

void immediate6502()
{
      savepc=PC++;
}

void abs6502()
{
      savepc = read6502(PC) + (read6502(PC + 1) << 8);
      PC++;
      PC++;
}

void relative6502()
{
      savepc = read6502(PC++);
      if (savepc & 0x80) savepc -= 0x100;
      if ((savepc>>8) != (PC>>8))
	      clockticks6502++;
}

void indirect6502()
{
      //if ((PC&0xFF)==0xFF) help = read6502(PC) + (read6502(PC&0xFF00) << 8);
        //else
        help = read6502(PC) + ((read6502(PC + 1)&0xFF) << 8);
      if ((help&0xFF)==0xFF) savepc = read6502(help) + (read6502(help&0xFF00) << 8);
        else savepc = read6502(help) + (read6502(help + 1) << 8);
      PC++;
      PC++;
}

void absx6502()
{
      savepc = read6502(PC) + (read6502(PC + 1) << 8);
      PC++;
      PC++;
      if (ticks[opcode]==4)
              if ((savepc>>8) != ((savepc+X)>>8))
		      clockticks6502++;
      savepc += X;
}

void absy6502()
{
      savepc = read6502(PC) + (read6502(PC + 1) << 8);
      PC++;
      PC++;

      if (ticks[opcode]==4)
              if ((savepc>>8) != ((savepc+Y)>>8))
		      clockticks6502++;
      savepc += Y;
}

void zp6502()
{
      savepc=read6502(PC++);
}

void zpx6502()
{
      savepc=read6502(PC++)+X;
      savepc &= 0x00ff;
}

void zpy6502()
{
      savepc=read6502(PC++)+Y;
      savepc &= 0x00ff;
}

void indx6502()
{
      value = read6502(PC++)+X;
      savepc = read6502(value) + (read6502(value+1) << 8);
}

void indy6502()
{
      value = read6502(PC++);
      savepc = read6502(value) + (read6502(value+1) << 8);
      if (ticks[opcode]==5)
              if ((savepc>>8) != ((savepc+Y)>>8))
		      clockticks6502++;
      savepc += Y;
}

void indabsx6502()
{
      help = read6502(PC) + (read6502(PC + 1) << 8) + X;
      savepc = read6502(help) + (read6502(help + 1) << 8);
}

void indzp6502()
{
      value = read6502(PC++);
      savepc = read6502(value) + (read6502(value + 1) << 8);
}

/* Instructions */
#define adc6502() { adrmode[opcode](); value = read6502(savepc); saveflags=(P & 0x01); sum= ((char) A) + ((char) value) + saveflags; if ((sum>0x7f) || (sum<-0x80)) P |= 0x40; else P &= 0xbf; sum= A + value + saveflags; if (sum>0xff) P |= 0x01; else P &= 0xfe; A=sum; if (P & 0x08) { P &= 0xfe; if ((A & 0x0f)>0x09) A += 0x06; if ((A & 0xf0)>0x90) { A += 0x60; P |= 0x01; } } else { clockticks6502++; } if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define and6502() { adrmode[opcode](); value = read6502(savepc); A &= value; if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define asl6502() { adrmode[opcode](); value = read6502(savepc); P= (P & 0xfe) | ((value >>7) & 0x01); value = value << 1; write6502(savepc,value); if (value) P &= 0xfd; else P |= 0x02; if (value & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define asla6502() { P= (P & 0xfe) | ((A >>7) & 0x01); A = A << 1; if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define bcc6502() { if ((P & 0x01)==0) { adrmode[opcode](); PC += savepc; clockticks6502++; } else value = read6502(PC++); } 
 #define bcs6502() { if (P & 0x01) { adrmode[opcode](); PC += savepc; clockticks6502++; } else value=read6502(PC++); } 
 #define beq6502() { if (P & 0x02) { adrmode[opcode](); PC += savepc; clockticks6502++; } else value=read6502(PC++); } 
 #define bit6502() { adrmode[opcode](); value=read6502(savepc);  if (value & A) P &= 0xfd; else P |= 0x02;  P = (P & 0x3f) | (value & 0xc0); } 
 #define bmi6502() { if (P & 0x80) { adrmode[opcode](); PC += savepc; clockticks6502++; } else value=read6502(PC++); } 
 #define bne6502() { if ((P & 0x02)==0) { adrmode[opcode](); PC += savepc; clockticks6502++; } else value=read6502(PC++); } 
 #define bpl6502() { if ((P & 0x80)==0) { adrmode[opcode](); PC += savepc; clockticks6502++; } else value=read6502(PC++); } 
 #define brk6502() { PC++; write6502(0x0100+S--,(BYTE)(PC>>8)); write6502(0x0100+S--,(BYTE)(PC & 0xff)); write6502(0x0100+S--,P); P |= 0x14; PC = read6502(0xfffe) + (read6502(0xffff) << 8); } 
 #define bvc6502() { if ((P & 0x40)==0) { adrmode[opcode](); PC += savepc; clockticks6502++; } else value=read6502(PC++); } 
 #define bvs6502() { if (P & 0x40) { adrmode[opcode](); PC += savepc; clockticks6502++; } else value=read6502(PC++); } 
 #define clc6502() { P &= 0xfe; } 
 #define cld6502() { } 
 #define cli6502() { P &= 0xfb; } 
 #define clv6502() { P &= 0xbf; } 
 #define cmp6502() { adrmode[opcode](); value = read6502(savepc); if (A+0x100-value>0xff) P |= 0x01; else P &= 0xfe; value=A+0x100-value; if (value) P &= 0xfd; else P |= 0x02; if (value & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define cpx6502() { adrmode[opcode](); value = read6502(savepc); if (X+0x100-value>0xff) P |= 0x01; else P &= 0xfe; value=X+0x100-value; if (value) P &= 0xfd; else P |= 0x02; if (value & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define cpy6502() { adrmode[opcode](); value = read6502(savepc); if (Y+0x100-value>0xff) P |= 0x01; else P &= 0xfe; value=Y+0x100-value; if (value) P &= 0xfd; else P |= 0x02; if (value & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define dec6502() { adrmode[opcode](); write6502(savepc, read6502(savepc) - 1); value = read6502(savepc); if (value) P &= 0xfd; else P |= 0x02; if (value & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define dex6502() { X--; if (X) P &= 0xfd; else P |= 0x02; if (X & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define dey6502() { Y--; if (Y) P &= 0xfd; else P |= 0x02; if (Y & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define eor6502() { adrmode[opcode](); A ^= read6502(savepc); if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define inc6502() { adrmode[opcode](); write6502(savepc, read6502(savepc) + 1); value = read6502(savepc); if (value) P &= 0xfd; else P |= 0x02; if (value & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define inx6502() { X++; if (X) P &= 0xfd; else P |= 0x02; if (X & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define iny6502() { Y++; if (Y) P &= 0xfd; else P |= 0x02; if (Y & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define jmp6502() { adrmode[opcode](); PC=savepc; } 
 #define jsr6502() { PC++; write6502(0x0100+S--,(BYTE)(PC >> 8)); write6502(0x0100+S--,(BYTE)(PC & 0xff)); PC--; adrmode[opcode](); PC=savepc; } 
 #define lda6502() { adrmode[opcode](); A=read6502(savepc); if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define ldx6502() { adrmode[opcode](); X=read6502(savepc); if (X) P &= 0xfd; else P |= 0x02; if (X & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define ldy6502() { adrmode[opcode](); Y=read6502(savepc); if (Y) P &= 0xfd; else P |= 0x02; if (Y & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define lsr6502() { adrmode[opcode](); value=read6502(savepc);  P= (P & 0xfe) | (value & 0x01);  value = value >>1; write6502(savepc,value);  if (value != 0) P &= 0xfd; else P |= 0x02;  if ((value & 0x80) == 0x80) P |= 0x80; else P &= 0x7f; } 
 #define lsra6502() { P= (P & 0xfe) | (A & 0x01); A = A >>1; if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define nop6502() { } 
 #define ora6502() { adrmode[opcode](); A |= read6502(savepc); if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define pha6502() { write6502(0x100+S--, A); } 
 #define php6502() { write6502(0x100+S--, P); } 
 #define pla6502() { A=read6502(++S+0x100); if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define plp6502() { P=read6502(++S+0x100) | 0x20; } 
 #define rol6502() { saveflags=(P & 0x01); adrmode[opcode](); value = read6502(savepc); P= (P & 0xfe) | ((value >>7) & 0x01); value = value << 1; value |= saveflags; write6502(savepc,value); if (value) P &= 0xfd; else P |= 0x02; if (value & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define rola6502() { saveflags=(P & 0x01); P= (P & 0xfe) | ((A >>7) & 0x01); A = A << 1; A |= saveflags; if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define ror6502() { saveflags=(P & 0x01); adrmode[opcode](); value=read6502(savepc); P= (P & 0xfe) | (value & 0x01); value = value >>1; if (saveflags) value |= 0x80; write6502(savepc,value); if (value) P &= 0xfd; else P |= 0x02; if (value & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define rora6502() { saveflags=(P & 0x01); P= (P & 0xfe) | (A & 0x01); A = A >>1; if (saveflags) A |= 0x80; if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define rti6502() { P=read6502(++S+0x100) | 0x20; PC=read6502(++S+0x100); PC |= (read6502(++S+0x100) << 8); } 
 #define rts6502() { PC=read6502(++S+0x100); PC |= (read6502(++S+0x100) << 8); PC++; } 
 #define sbc6502() { adrmode[opcode](); value = read6502(savepc) ^ 0xff; saveflags=(P & 0x01); sum= ((char) A) + ((char) value) + (saveflags << 4); if ((sum>0x7f) || (sum<-0x80)) P |= 0x40; else P &= 0xbf; sum= A + value + saveflags; if (sum>0xff) P |= 0x01; else P &= 0xfe; A=sum; if (P & 0x08) { A -= 0x66; P &= 0xfe; if ((A & 0x0f)>0x09) A += 0x06; if ((A & 0xf0)>0x90) { A += 0x60; P |= 0x01; } } else { clockticks6502++; } if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define sec6502() { P |= 0x01; } 
 #define sed6502() { } 
 #define sei6502() { P |= 0x04; } 
 #define sta6502() { adrmode[opcode](); write6502(savepc,A); } 
 #define stx6502() { adrmode[opcode](); write6502(savepc,X); } 
 #define sty6502() { adrmode[opcode](); write6502(savepc,Y); } 
 #define tax6502() { X=A; if (X) P &= 0xfd; else P |= 0x02; if (X & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define tay6502() { Y=A; if (Y) P &= 0xfd; else P |= 0x02; if (Y & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define tsx6502() { X=S; if (X) P &= 0xfd; else P |= 0x02; if (X & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define txa6502() { A=X; if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define txs6502() { S=X; } 
 #define tya6502() { A=Y; if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define bra6502() { adrmode[opcode](); PC += savepc; clockticks6502++; } 
 #define dea6502() { A--; if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define ina6502() { A++; if (A) P &= 0xfd; else P |= 0x02; if (A & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define phx6502() { write6502(0x100+S--,X); } 
 #define plx6502() { X=read6502(++S+0x100); if (X) P &= 0xfd; else P |= 0x02; if (X & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define phy6502() { write6502(0x100+S--,Y); } 
 #define ply6502() { Y=read6502(++S+0x100); if (Y) P &= 0xfd; else P |= 0x02; if (Y & 0x80) P |= 0x80; else P &= 0x7f; } 
 #define stz6502() { adrmode[opcode](); write6502(savepc,0); } 
 #define tsb6502() { adrmode[opcode](); write6502(savepc, read6502(savepc) | A); if (read6502(savepc)) P &= 0xfd; else P |= 0x02; } 
 #define trb6502() { adrmode[opcode](); write6502(savepc, read6502(savepc) & (A ^ 0xff)); if (read6502(savepc)) P &= 0xfd; else P |= 0x02; } 

/* Reset CPU */
void reset6502()
{
       A=X=Y=P=0;
       P |= 0x20;
       S=0xff;
       PC=read6502(0xfffc);
       PC |= read6502(0xfffd) << 8;
}

/* Init CPU */
void init6502()
{
      ticks[0x00]=7; adrmode[0x00]=implied6502;
      ticks[0x01]=6; adrmode[0x01]=indx6502;
      ticks[0x02]=2; adrmode[0x02]=implied6502;
      ticks[0x03]=2; adrmode[0x03]=implied6502;
      ticks[0x04]=3; adrmode[0x04]=zp6502;
      ticks[0x05]=3; adrmode[0x05]=zp6502;
      ticks[0x06]=5; adrmode[0x06]=zp6502;
      ticks[0x07]=2; adrmode[0x07]=implied6502;
      ticks[0x08]=3; adrmode[0x08]=implied6502;
      ticks[0x09]=2; adrmode[0x09]=immediate6502;
      ticks[0x0a]=2; adrmode[0x0a]=implied6502;
      ticks[0x0b]=2; adrmode[0x0b]=implied6502;
      ticks[0x0c]=4; adrmode[0x0c]=abs6502;
      ticks[0x0d]=4; adrmode[0x0d]=abs6502;
      ticks[0x0e]=6; adrmode[0x0e]=abs6502;
      ticks[0x0f]=2; adrmode[0x0f]=implied6502;
      ticks[0x10]=2; adrmode[0x10]=relative6502;
      ticks[0x11]=5; adrmode[0x11]=indy6502;
      ticks[0x12]=3; adrmode[0x12]=indzp6502;
      ticks[0x13]=2; adrmode[0x13]=implied6502;
      ticks[0x14]=3; adrmode[0x14]=zp6502;
      ticks[0x15]=4; adrmode[0x15]=zpx6502;
      ticks[0x16]=6; adrmode[0x16]=zpx6502;
      ticks[0x17]=2; adrmode[0x17]=implied6502;
      ticks[0x18]=2; adrmode[0x18]=implied6502;
      ticks[0x19]=4; adrmode[0x19]=absy6502;
      ticks[0x1a]=2; adrmode[0x1a]=implied6502;
      ticks[0x1b]=2; adrmode[0x1b]=implied6502;
      ticks[0x1c]=4; adrmode[0x1c]=abs6502;
      ticks[0x1d]=4; adrmode[0x1d]=absx6502;
      ticks[0x1e]=7; adrmode[0x1e]=absx6502;
      ticks[0x1f]=2; adrmode[0x1f]=implied6502;
      ticks[0x20]=6; adrmode[0x20]=abs6502;
      ticks[0x21]=6; adrmode[0x21]=indx6502;
      ticks[0x22]=2; adrmode[0x22]=implied6502;
      ticks[0x23]=2; adrmode[0x23]=implied6502;
      ticks[0x24]=3; adrmode[0x24]=zp6502;
      ticks[0x25]=3; adrmode[0x25]=zp6502;
      ticks[0x26]=5; adrmode[0x26]=zp6502;
      ticks[0x27]=2; adrmode[0x27]=implied6502;
      ticks[0x28]=4; adrmode[0x28]=implied6502;
      ticks[0x29]=2; adrmode[0x29]=immediate6502;
      ticks[0x2a]=2; adrmode[0x2a]=implied6502;
      ticks[0x2b]=2; adrmode[0x2b]=implied6502;
      ticks[0x2c]=4; adrmode[0x2c]=abs6502;
      ticks[0x2d]=4; adrmode[0x2d]=abs6502;
      ticks[0x2e]=6; adrmode[0x2e]=abs6502;
      ticks[0x2f]=2; adrmode[0x2f]=implied6502;
      ticks[0x30]=2; adrmode[0x30]=relative6502;
      ticks[0x31]=5; adrmode[0x31]=indy6502;
      ticks[0x32]=3; adrmode[0x32]=indzp6502;
      ticks[0x33]=2; adrmode[0x33]=implied6502;
      ticks[0x34]=4; adrmode[0x34]=zpx6502;
      ticks[0x35]=4; adrmode[0x35]=zpx6502;
      ticks[0x36]=6; adrmode[0x36]=zpx6502;
      ticks[0x37]=2; adrmode[0x37]=implied6502;
      ticks[0x38]=2; adrmode[0x38]=implied6502;
      ticks[0x39]=4; adrmode[0x39]=absy6502;
      ticks[0x3a]=2; adrmode[0x3a]=implied6502;
      ticks[0x3b]=2; adrmode[0x3b]=implied6502;
      ticks[0x3c]=4; adrmode[0x3c]=absx6502;
      ticks[0x3d]=4; adrmode[0x3d]=absx6502;
      ticks[0x3e]=7; adrmode[0x3e]=absx6502;
      ticks[0x3f]=2; adrmode[0x3f]=implied6502;
      ticks[0x40]=6; adrmode[0x40]=implied6502;
      ticks[0x41]=6; adrmode[0x41]=indx6502;
      ticks[0x42]=2; adrmode[0x42]=implied6502;
      ticks[0x43]=2; adrmode[0x43]=implied6502;
      ticks[0x44]=2; adrmode[0x44]=implied6502;
      ticks[0x45]=3; adrmode[0x45]=zp6502;
      ticks[0x46]=5; adrmode[0x46]=zp6502;
      ticks[0x47]=2; adrmode[0x47]=implied6502;
      ticks[0x48]=3; adrmode[0x48]=implied6502;
      ticks[0x49]=2; adrmode[0x49]=immediate6502;
      ticks[0x4a]=2; adrmode[0x4a]=implied6502;
      ticks[0x4b]=2; adrmode[0x4b]=implied6502;
      ticks[0x4c]=3; adrmode[0x4c]=abs6502;
      ticks[0x4d]=4; adrmode[0x4d]=abs6502;
      ticks[0x4e]=6; adrmode[0x4e]=abs6502;
      ticks[0x4f]=2; adrmode[0x4f]=implied6502;
      ticks[0x50]=2; adrmode[0x50]=relative6502;
      ticks[0x51]=5; adrmode[0x51]=indy6502;
      ticks[0x52]=3; adrmode[0x52]=indzp6502;
      ticks[0x53]=2; adrmode[0x53]=implied6502;
      ticks[0x54]=2; adrmode[0x54]=implied6502;
      ticks[0x55]=4; adrmode[0x55]=zpx6502;
      ticks[0x56]=6; adrmode[0x56]=zpx6502;
      ticks[0x57]=2; adrmode[0x57]=implied6502;
      ticks[0x58]=2; adrmode[0x58]=implied6502;
      ticks[0x59]=4; adrmode[0x59]=absy6502;
      ticks[0x5a]=3; adrmode[0x5a]=implied6502;
      ticks[0x5b]=2; adrmode[0x5b]=implied6502;
      ticks[0x5c]=2; adrmode[0x5c]=implied6502;
      ticks[0x5d]=4; adrmode[0x5d]=absx6502;
      ticks[0x5e]=7; adrmode[0x5e]=absx6502;
      ticks[0x5f]=2; adrmode[0x5f]=implied6502;
      ticks[0x60]=6; adrmode[0x60]=implied6502;
      ticks[0x61]=6; adrmode[0x61]=indx6502;
      ticks[0x62]=2; adrmode[0x62]=implied6502;
      ticks[0x63]=2; adrmode[0x63]=implied6502;
      ticks[0x64]=3; adrmode[0x64]=zp6502;
      ticks[0x65]=3; adrmode[0x65]=zp6502;
      ticks[0x66]=5; adrmode[0x66]=zp6502;
      ticks[0x67]=2; adrmode[0x67]=implied6502;
      ticks[0x68]=4; adrmode[0x68]=implied6502;
      ticks[0x69]=2; adrmode[0x69]=immediate6502;
      ticks[0x6a]=2; adrmode[0x6a]=implied6502;
      ticks[0x6b]=2; adrmode[0x6b]=implied6502;
      ticks[0x6c]=5; adrmode[0x6c]=indirect6502;
      ticks[0x6d]=4; adrmode[0x6d]=abs6502;
      ticks[0x6e]=6; adrmode[0x6e]=abs6502;
      ticks[0x6f]=2; adrmode[0x6f]=implied6502;
      ticks[0x70]=2; adrmode[0x70]=relative6502;
      ticks[0x71]=5; adrmode[0x71]=indy6502;
      ticks[0x72]=3; adrmode[0x72]=indzp6502;
      ticks[0x73]=2; adrmode[0x73]=implied6502;
      ticks[0x74]=4; adrmode[0x74]=zpx6502;
      ticks[0x75]=4; adrmode[0x75]=zpx6502;
      ticks[0x76]=6; adrmode[0x76]=zpx6502;
      ticks[0x77]=2; adrmode[0x77]=implied6502;
      ticks[0x78]=2; adrmode[0x78]=implied6502;
      ticks[0x79]=4; adrmode[0x79]=absy6502;
      ticks[0x7a]=4; adrmode[0x7a]=implied6502;
      ticks[0x7b]=2; adrmode[0x7b]=implied6502;
      ticks[0x7c]=6; adrmode[0x7c]=indabsx6502;
      ticks[0x7d]=4; adrmode[0x7d]=absx6502;
      ticks[0x7e]=7; adrmode[0x7e]=absx6502;
      ticks[0x7f]=2; adrmode[0x7f]=implied6502;
      ticks[0x80]=2; adrmode[0x80]=relative6502;
      ticks[0x81]=6; adrmode[0x81]=indx6502;
      ticks[0x82]=2; adrmode[0x82]=implied6502;
      ticks[0x83]=2; adrmode[0x83]=implied6502;
      ticks[0x84]=3; adrmode[0x84]=zp6502;
      ticks[0x85]=3; adrmode[0x85]=zp6502;
      ticks[0x86]=3; adrmode[0x86]=zp6502;
      ticks[0x87]=2; adrmode[0x87]=implied6502;
      ticks[0x88]=2; adrmode[0x88]=implied6502;
      ticks[0x89]=2; adrmode[0x89]=immediate6502;
      ticks[0x8a]=2; adrmode[0x8a]=implied6502;
      ticks[0x8b]=2; adrmode[0x8b]=implied6502;
      ticks[0x8c]=4; adrmode[0x8c]=abs6502;
      ticks[0x8d]=4; adrmode[0x8d]=abs6502;
      ticks[0x8e]=4; adrmode[0x8e]=abs6502;
      ticks[0x8f]=2; adrmode[0x8f]=implied6502;
      ticks[0x90]=2; adrmode[0x90]=relative6502;
      ticks[0x91]=6; adrmode[0x91]=indy6502;
      ticks[0x92]=3; adrmode[0x92]=indzp6502;
      ticks[0x93]=2; adrmode[0x93]=implied6502;
      ticks[0x94]=4; adrmode[0x94]=zpx6502;
      ticks[0x95]=4; adrmode[0x95]=zpx6502;
      ticks[0x96]=4; adrmode[0x96]=zpy6502;
      ticks[0x97]=2; adrmode[0x97]=implied6502;
      ticks[0x98]=2; adrmode[0x98]=implied6502;
      ticks[0x99]=5; adrmode[0x99]=absy6502;
      ticks[0x9a]=2; adrmode[0x9a]=implied6502;
      ticks[0x9b]=2; adrmode[0x9b]=implied6502;
      ticks[0x9c]=4; adrmode[0x9c]=abs6502;
      ticks[0x9d]=5; adrmode[0x9d]=absx6502;
      ticks[0x9e]=5; adrmode[0x9e]=absx6502;
      ticks[0x9f]=2; adrmode[0x9f]=implied6502;
      ticks[0xa0]=2; adrmode[0xa0]=immediate6502;
      ticks[0xa1]=6; adrmode[0xa1]=indx6502;
      ticks[0xa2]=2; adrmode[0xa2]=immediate6502;
      ticks[0xa3]=2; adrmode[0xa3]=implied6502;
      ticks[0xa4]=3; adrmode[0xa4]=zp6502;
      ticks[0xa5]=3; adrmode[0xa5]=zp6502;
      ticks[0xa6]=3; adrmode[0xa6]=zp6502;
      ticks[0xa7]=2; adrmode[0xa7]=implied6502;
      ticks[0xa8]=2; adrmode[0xa8]=implied6502;
      ticks[0xa9]=2; adrmode[0xa9]=immediate6502;
      ticks[0xaa]=2; adrmode[0xaa]=implied6502;
      ticks[0xab]=2; adrmode[0xab]=implied6502;
      ticks[0xac]=4; adrmode[0xac]=abs6502;
      ticks[0xad]=4; adrmode[0xad]=abs6502;
      ticks[0xae]=4; adrmode[0xae]=abs6502;
      ticks[0xaf]=2; adrmode[0xaf]=implied6502;
      ticks[0xb0]=2; adrmode[0xb0]=relative6502;
      ticks[0xb1]=5; adrmode[0xb1]=indy6502;
      ticks[0xb2]=3; adrmode[0xb2]=indzp6502;
      ticks[0xb3]=2; adrmode[0xb3]=implied6502;
      ticks[0xb4]=4; adrmode[0xb4]=zpx6502;
      ticks[0xb5]=4; adrmode[0xb5]=zpx6502;
      ticks[0xb6]=4; adrmode[0xb6]=zpy6502;
      ticks[0xb7]=2; adrmode[0xb7]=implied6502;
      ticks[0xb8]=2; adrmode[0xb8]=implied6502;
      ticks[0xb9]=4; adrmode[0xb9]=absy6502;
      ticks[0xba]=2; adrmode[0xba]=implied6502;
      ticks[0xbb]=2; adrmode[0xbb]=implied6502;
      ticks[0xbc]=4; adrmode[0xbc]=absx6502;
      ticks[0xbd]=4; adrmode[0xbd]=absx6502;
      ticks[0xbe]=4; adrmode[0xbe]=absy6502;
      ticks[0xbf]=2; adrmode[0xbf]=implied6502;
      ticks[0xc0]=2; adrmode[0xc0]=immediate6502;
      ticks[0xc1]=6; adrmode[0xc1]=indx6502;
      ticks[0xc2]=2; adrmode[0xc2]=implied6502;
      ticks[0xc3]=2; adrmode[0xc3]=implied6502;
      ticks[0xc4]=3; adrmode[0xc4]=zp6502;
      ticks[0xc5]=3; adrmode[0xc5]=zp6502;
      ticks[0xc6]=5; adrmode[0xc6]=zp6502;
      ticks[0xc7]=2; adrmode[0xc7]=implied6502;
      ticks[0xc8]=2; adrmode[0xc8]=implied6502;
      ticks[0xc9]=2; adrmode[0xc9]=immediate6502;
      ticks[0xca]=2; adrmode[0xca]=implied6502;
      ticks[0xcb]=2; adrmode[0xcb]=implied6502;
      ticks[0xcc]=4; adrmode[0xcc]=abs6502;
      ticks[0xcd]=4; adrmode[0xcd]=abs6502;
      ticks[0xce]=6; adrmode[0xce]=abs6502;
      ticks[0xcf]=2; adrmode[0xcf]=implied6502;
      ticks[0xd0]=2; adrmode[0xd0]=relative6502;
      ticks[0xd1]=5; adrmode[0xd1]=indy6502;
      ticks[0xd2]=3; adrmode[0xd2]=indzp6502;
      ticks[0xd3]=2; adrmode[0xd3]=implied6502;
      ticks[0xd4]=2; adrmode[0xd4]=implied6502;
      ticks[0xd5]=4; adrmode[0xd5]=zpx6502;
      ticks[0xd6]=6; adrmode[0xd6]=zpx6502;
      ticks[0xd7]=2; adrmode[0xd7]=implied6502;
      ticks[0xd8]=2; adrmode[0xd8]=implied6502;
      ticks[0xd9]=4; adrmode[0xd9]=absy6502;
      ticks[0xda]=3; adrmode[0xda]=implied6502;
      ticks[0xdb]=2; adrmode[0xdb]=implied6502;
      ticks[0xdc]=2; adrmode[0xdc]=implied6502;
      ticks[0xdd]=4; adrmode[0xdd]=absx6502;
      ticks[0xde]=7; adrmode[0xde]=absx6502;
      ticks[0xdf]=2; adrmode[0xdf]=implied6502;
      ticks[0xe0]=2; adrmode[0xe0]=immediate6502;
      ticks[0xe1]=6; adrmode[0xe1]=indx6502;
      ticks[0xe2]=2; adrmode[0xe2]=implied6502;
      ticks[0xe3]=2; adrmode[0xe3]=implied6502;
      ticks[0xe4]=3; adrmode[0xe4]=zp6502;
      ticks[0xe5]=3; adrmode[0xe5]=zp6502;
      ticks[0xe6]=5; adrmode[0xe6]=zp6502;
      ticks[0xe7]=2; adrmode[0xe7]=implied6502;
      ticks[0xe8]=2; adrmode[0xe8]=implied6502;
      ticks[0xe9]=2; adrmode[0xe9]=immediate6502;
      ticks[0xea]=2; adrmode[0xea]=implied6502;
      ticks[0xeb]=2; adrmode[0xeb]=implied6502;
      ticks[0xec]=4; adrmode[0xec]=abs6502;
      ticks[0xed]=4; adrmode[0xed]=abs6502;
      ticks[0xee]=6; adrmode[0xee]=abs6502;
      ticks[0xef]=2; adrmode[0xef]=implied6502;
      ticks[0xf0]=2; adrmode[0xf0]=relative6502;
      ticks[0xf1]=5; adrmode[0xf1]=indy6502;
      ticks[0xf2]=3; adrmode[0xf2]=indzp6502;
      ticks[0xf3]=2; adrmode[0xf3]=implied6502;
      ticks[0xf4]=2; adrmode[0xf4]=implied6502;
      ticks[0xf5]=4; adrmode[0xf5]=zpx6502;
      ticks[0xf6]=6; adrmode[0xf6]=zpx6502;
      ticks[0xf7]=2; adrmode[0xf7]=implied6502;
      ticks[0xf8]=2; adrmode[0xf8]=implied6502;
      ticks[0xf9]=4; adrmode[0xf9]=absy6502;
      ticks[0xfa]=4; adrmode[0xfa]=implied6502;
      ticks[0xfb]=2; adrmode[0xfb]=implied6502;
      ticks[0xfc]=2; adrmode[0xfc]=implied6502;
      ticks[0xfd]=4; adrmode[0xfd]=absx6502;
      ticks[0xfe]=7; adrmode[0xfe]=absx6502;
      ticks[0xff]=2; adrmode[0xff]=implied6502;
//      sei6502();
}

/* Non maskerable interrupt */
void nmi6502()
{
      write6502(0x0100+S--,(BYTE)(PC>>8));
      write6502(0x0100+S--,(BYTE)(PC & 0xff));
      write6502(0x0100+S--,P);
      P |= 0x04;
      PC=read6502(0xfffa);
      PC |= read6502(0xfffb) << 8;
}

/* Maskerable Interrupt */
void irq6502()
{
   if (P & 0x04) return;
   write6502(0x0100+S--,(BYTE)(PC>>8));
   write6502(0x0100+S--,(BYTE)(PC & 0xff));
   write6502(0x0100+S--,P);
   P |= 0x04;
   PC=read6502(0xfffe);
   PC |= read6502(0xffff) << 8;
}

#ifdef DEBUG_DUMP
	extern void dasm(unsigned op,
		unsigned PC,
		unsigned A,unsigned X, unsigned Y,unsigned Flags,unsigned SP,
		unsigned PPU_X, unsigned PPU_Scanline);
	uint16_t lastop;
#endif

/* Execute Instruction */
extern int buflen, bufpos;
extern void tickchannelsAPU();
uint16_t lastop;
uint32_t clockgoal6502 = 0;
uint32_t exec6502(uint32_t timerTicks)
{
	uint32_t startticks;
	clockgoal6502 += timerTicks;
	startticks = clockticks6502;

   //while (timerTicks + remainTicks > 0)
	while (clockticks6502 < clockgoal6502)
   {
      //if (tracelog==1) printf("%x: ", PC);
      // fetch instruction
      opcode = read6502(PC++);
	  lastop = PC;
      //if (tracelog==1) printf("%x\n", opcode);
      //if (tracelog==1 && opcode==0xad) { printf("AD absolute: %x\n", read6502(PC)); printf("AD absolute 2: %x\n", read6502(PC+1)); }
      //if (tracelog==1 && opcode==0x29) printf("val: %x\n", read6502(PC));
      // execute instruction
      //instruction[opcode]();
      switch (opcode) {
		case 0x00: brk6502(); break;
		case 0x01: ora6502(); break;
		case 0x02: nop6502(); break;
		case 0x03: nop6502(); break;
		case 0x04: tsb6502(); break;
		case 0x05: ora6502(); break;
		case 0x06: asl6502(); break;
		case 0x07: nop6502(); break;
		case 0x08: php6502(); break;
		case 0x09: ora6502(); break;
		case 0x0a: asla6502(); break;
		case 0x0b: nop6502(); break;
		case 0x0c: tsb6502(); break;
		case 0x0d: ora6502(); break;
		case 0x0e: asl6502(); break;
		case 0x0f: nop6502(); break;
		case 0x10: bpl6502(); break;
		case 0x11: ora6502(); break;
		case 0x12: ora6502(); break;
		case 0x13: nop6502(); break;
		case 0x14: trb6502(); break;
		case 0x15: ora6502(); break;
		case 0x16: asl6502(); break;
		case 0x17: nop6502(); break;
		case 0x18: clc6502(); break;
		case 0x19: ora6502(); break;
		case 0x1a: ina6502(); break;
		case 0x1b: nop6502(); break;
		case 0x1c: trb6502(); break;
		case 0x1d: ora6502(); break;
		case 0x1e: asl6502(); break;
		case 0x1f: nop6502(); break;
		case 0x20: jsr6502(); break;
		case 0x21: and6502(); break;
		case 0x22: nop6502(); break;
		case 0x23: nop6502(); break;
		case 0x24: bit6502(); break;
		case 0x25: and6502(); break;
		case 0x26: rol6502(); break;
		case 0x27: nop6502(); break;
		case 0x28: plp6502(); break;
		case 0x29: and6502(); break;
		case 0x2a: rola6502(); break;
		case 0x2b: nop6502(); break;
		case 0x2c: bit6502(); break;
		case 0x2d: and6502(); break;
		case 0x2e: rol6502(); break;
		case 0x2f: nop6502(); break;
		case 0x30: bmi6502(); break;
		case 0x31: and6502(); break;
		case 0x32: and6502(); break;
		case 0x33: nop6502(); break;
		case 0x34: bit6502(); break;
		case 0x35: and6502(); break;
		case 0x36: rol6502(); break;
		case 0x37: nop6502(); break;
		case 0x38: sec6502(); break;
		case 0x39: and6502(); break;
		case 0x3a: dea6502(); break;
		case 0x3b: nop6502(); break;
		case 0x3c: bit6502(); break;
		case 0x3d: and6502(); break;
		case 0x3e: rol6502(); break;
		case 0x3f: nop6502(); break;
		case 0x40: rti6502(); break;
		case 0x41: eor6502(); break;
		case 0x42: nop6502(); break;
		case 0x43: nop6502(); break;
		case 0x44: nop6502(); break;
		case 0x45: eor6502(); break;
		case 0x46: lsr6502(); break;
		case 0x47: nop6502(); break;
		case 0x48: pha6502(); break;
		case 0x49: eor6502(); break;
		case 0x4a: lsra6502(); break;
		case 0x4b: nop6502(); break;
		case 0x4c: jmp6502(); break;
		case 0x4d: eor6502(); break;
		case 0x4e: lsr6502(); break;
		case 0x4f: nop6502(); break;
		case 0x50: bvc6502(); break;
		case 0x51: eor6502(); break;
		case 0x52: eor6502(); break;
		case 0x53: nop6502(); break;
		case 0x54: nop6502(); break;
		case 0x55: eor6502(); break;
		case 0x56: lsr6502(); break;
		case 0x57: nop6502(); break;
		case 0x58: cli6502(); break;
		case 0x59: eor6502(); break;
		case 0x5a: phy6502(); break;
		case 0x5b: nop6502(); break;
		case 0x5c: nop6502(); break;
		case 0x5d: eor6502(); break;
		case 0x5e: lsr6502(); break;
		case 0x5f: nop6502(); break;
		case 0x60: rts6502(); break;
		case 0x61: adc6502(); break;
		case 0x62: nop6502(); break;
		case 0x63: nop6502(); break;
		case 0x64: stz6502(); break;
		case 0x65: adc6502(); break;
		case 0x66: ror6502(); break;
		case 0x67: nop6502(); break;
		case 0x68: pla6502(); break;
		case 0x69: adc6502(); break;
		case 0x6a: rora6502(); break;
		case 0x6b: nop6502(); break;
		case 0x6c: jmp6502(); break;
		case 0x6d: adc6502(); break;
		case 0x6e: ror6502(); break;
		case 0x6f: nop6502(); break;
		case 0x70: bvs6502(); break;
		case 0x71: adc6502(); break;
		case 0x72: adc6502(); break;
		case 0x73: nop6502(); break;
		case 0x74: stz6502(); break;
		case 0x75: adc6502(); break;
		case 0x76: ror6502(); break;
		case 0x77: nop6502(); break;
		case 0x78: sei6502(); break;
		case 0x79: adc6502(); break;
		case 0x7a: ply6502(); break;
		case 0x7b: nop6502(); break;
		case 0x7c: jmp6502(); break;
		case 0x7d: adc6502(); break;
		case 0x7e: ror6502(); break;
		case 0x7f: nop6502(); break;
		case 0x80: bra6502(); break;
		case 0x81: sta6502(); break;
		case 0x82: nop6502(); break;
		case 0x83: nop6502(); break;
		case 0x84: sty6502(); break;
		case 0x85: sta6502(); break;
		case 0x86: stx6502(); break;
		case 0x87: nop6502(); break;
		case 0x88: dey6502(); break;
		case 0x89: bit6502(); break;
		case 0x8a: txa6502(); break;
		case 0x8b: nop6502(); break;
		case 0x8c: sty6502(); break;
		case 0x8d: sta6502(); break;
		case 0x8e: stx6502(); break;
		case 0x8f: nop6502(); break;
		case 0x90: bcc6502(); break;
		case 0x91: sta6502(); break;
		case 0x92: sta6502(); break;
		case 0x93: nop6502(); break;
		case 0x94: sty6502(); break;
		case 0x95: sta6502(); break;
		case 0x96: stx6502(); break;
		case 0x97: nop6502(); break;
		case 0x98: tya6502(); break;
		case 0x99: sta6502(); break;
		case 0x9a: txs6502(); break;
		case 0x9b: nop6502(); break;
		case 0x9c: stz6502(); break;
		case 0x9d: sta6502(); break;
		case 0x9e: stz6502(); break;
		case 0x9f: nop6502(); break;
		case 0xa0: ldy6502(); break;
		case 0xa1: lda6502(); break;
		case 0xa2: ldx6502(); break;
		case 0xa3: nop6502(); break;
		case 0xa4: ldy6502(); break;
		case 0xa5: lda6502(); break;
		case 0xa6: ldx6502(); break;
		case 0xa7: nop6502(); break;
		case 0xa8: tay6502(); break;
		case 0xa9: lda6502(); break;
		case 0xaa: tax6502(); break;
		case 0xab: nop6502(); break;
		case 0xac: ldy6502(); break;
		case 0xad: lda6502(); break;
		case 0xae: ldx6502(); break;
		case 0xaf: nop6502(); break;
		case 0xb0: bcs6502(); break;
		case 0xb1: lda6502(); break;
		case 0xb2: lda6502(); break;
		case 0xb3: nop6502(); break;
		case 0xb4: ldy6502(); break;
		case 0xb5: lda6502(); break;
		case 0xb6: ldx6502(); break;
		case 0xb7: nop6502(); break;
		case 0xb8: clv6502(); break;
		case 0xb9: lda6502(); break;
		case 0xba: tsx6502(); break;
		case 0xbb: nop6502(); break;
		case 0xbc: ldy6502(); break;
		case 0xbd: lda6502(); break;
		case 0xbe: ldx6502(); break;
		case 0xbf: nop6502(); break;
		case 0xc0: cpy6502(); break;
		case 0xc1: cmp6502(); break;
		case 0xc2: nop6502(); break;
		case 0xc3: nop6502(); break;
		case 0xc4: cpy6502(); break;
		case 0xc5: cmp6502(); break;
		case 0xc6: dec6502(); break;
		case 0xc7: nop6502(); break;
		case 0xc8: iny6502(); break;
		case 0xc9: cmp6502(); break;
		case 0xca: dex6502(); break;
		case 0xcb: nop6502(); break;
		case 0xcc: cpy6502(); break;
		case 0xcd: cmp6502(); break;
		case 0xce: dec6502(); break;
		case 0xcf: nop6502(); break;
		case 0xd0: bne6502(); break;
		case 0xd1: cmp6502(); break;
		case 0xd2: cmp6502(); break;
		case 0xd3: nop6502(); break;
		case 0xd4: nop6502(); break;
		case 0xd5: cmp6502(); break;
		case 0xd6: dec6502(); break;
		case 0xd7: nop6502(); break;
		case 0xd8: cld6502(); break;
		case 0xd9: cmp6502(); break;
		case 0xda: phx6502(); break;
		case 0xdb: nop6502(); break;
		case 0xdc: nop6502(); break;
		case 0xdd: cmp6502(); break;
		case 0xde: dec6502(); break;
		case 0xdf: nop6502(); break;
		case 0xe0: cpx6502(); break;
		case 0xe1: sbc6502(); break;
		case 0xe2: nop6502(); break;
		case 0xe3: nop6502(); break;
		case 0xe4: cpx6502(); break;
		case 0xe5: sbc6502(); break;
		case 0xe6: inc6502(); break;
		case 0xe7: nop6502(); break;
		case 0xe8: inx6502(); break;
		case 0xe9: sbc6502(); break;
		case 0xea: nop6502(); break;
		case 0xeb: nop6502(); break;
		case 0xec: cpx6502(); break;
		case 0xed: sbc6502(); break;
		case 0xee: inc6502(); break;
		case 0xef: nop6502(); break;
		case 0xf0: beq6502(); break;
		case 0xf1: sbc6502(); break;
		case 0xf2: sbc6502(); break;
		case 0xf3: nop6502(); break;
		case 0xf4: nop6502(); break;
		case 0xf5: sbc6502(); break;
		case 0xf6: inc6502(); break;
		case 0xf7: nop6502(); break;
		case 0xf8: sed6502(); break;
		case 0xf9: sbc6502(); break;
		case 0xfa: plx6502(); break;
		case 0xfb: nop6502(); break;
		case 0xfc: nop6502(); break;
		case 0xfd: sbc6502(); break;
		case 0xfe: inc6502(); break;
		case 0xff: nop6502(); break;
		default: break;
      }
	#ifdef DEBUG_DUMP
		dasm(opcode, lastop, a_reg, x_reg, y_reg, flag_reg, s_reg, 0, 0);
	#endif

      // calculate clock cycles
      clockticks6502 += ticks[opcode];
      timerTicks -= ticks[opcode]; //clockticks6502;

	  tickchannelsAPU();
      }
   //}
//   remainTicks = timerTicks;
	  return(clockticks6502 - startticks);
}

uint16_t getPC() {
	return(PC);
}

uint32_t getclockticks() {
	return(clockticks6502);
}
#endif