#include <stdio.h>
#include <malloc.h>
#include "common.h"
#include "stdint.h"
#include "rom.h"
#include "PPU.h"

uint8_t gamegenie = 0;
uint8_t rompath[MAX_PATH];
uint8_t romloaded = 0;
struct cart_s cartridge;
extern struct PPU_s *PPU;
static const uint8_t validhdr[4] = { 'N', 'E', 'S', 0x1A };
extern void fatalerr(uint8_t *errmsg);
extern bool mapperinit(struct cart_s *cart);
extern uint8_t app_path[MAX_PATH];
extern int16_t msgbox(uint8_t *msgtxt, uint16_t buttons);

bool loadROM(uint8_t *filename) {
	FILE *romfile = NULL;
	static struct hdr_s header;
	uint16_t setptr;
	uint8_t path[MAX_PATH];

	if (gamegenie == 0) {
		romfile = fopen(filename, "rb");
		sprintf(&rompath[0], filename);
	} else if (gamegenie == 1) {
		sprintf(&path[0], "%sgenie.nes", &app_path[0]);
		romfile = fopen(&path[0], "rb");
		if (romfile == NULL) {
			msgbox("Unable to open %s as Game Genie ROM.", MB_ICONERROR | MB_OK);
			gamegenie = 0;
			romfile = fopen(filename, "rb");
		} else sprintf(&rompath[0], filename);
	} else if (gamegenie == 2) {
		romfile = fopen(&rompath[0], "rb");
	}
	romloaded = 1;

	fread(&header, 1, sizeof(header), romfile);
	if (memcmp((void *)&header.match, (void *)&validhdr[0], 4)) return(false);

	//printf("ROM information:\n");
	//parse through the header data
	if (header.flags[1] == 'D') header.flags[1] = 0; //DISKDUDE fix
	cartridge.mapper = ((header.flags[0] >> 4) | (header.flags[1] & 0xF0)) & 31;
	cartridge.hasSRAM = 1; //(header.flags[0] & 2) >> 1;
	if (header.flags[0] & 8) cartridge.mirroring = MIR_FOURSCREEN;
		else if (header.flags[0] & 1) cartridge.mirroring = MIR_VERTICAL;
			else cartridge.mirroring = MIR_HORIZONTAL;
	if (header.flags[0] & 4) cartridge.hasTrainer = true;
		else cartridge.hasTrainer = false;
	cartridge.PRGcount = (uint16_t)header.PRGsize << 4; //header gives PRG size in 16 KB units, we want it in 1 KB
	cartridge.CHRcount = (uint16_t)header.CHRsize << 3; //and CHR size is in 8 KB units

	switch (cartridge.mirroring) {
		case 0:
			PPU->ntmap[0] = 0x2000;
			PPU->ntmap[1] = 0x2000;
			PPU->ntmap[2] = 0x2400;
			PPU->ntmap[3] = 0x2400;
			break;
		case 1:
			PPU->ntmap[0] = 0x2000;
			PPU->ntmap[1] = 0x2400;
			PPU->ntmap[2] = 0x2000;
			PPU->ntmap[3] = 0x2400;
			break;
		case 2:
			PPU->ntmap[0] = 0x2000;
			PPU->ntmap[1] = 0x2400;
			PPU->ntmap[2] = 0x2600;
			PPU->ntmap[3] = 0x2800;
			break;
		default:
			break;
	}

	cartridge.PRGfull = malloc(cartridge.PRGcount << 10);
	if (cartridge.PRGfull == NULL) fatalerr("Unable to allocate for PRG-ROM data!");
	if (cartridge.CHRcount > 0) {
		cartridge.CHRfull = malloc(cartridge.CHRcount << 10);
		if (cartridge.CHRfull == NULL) fatalerr("Unable to allocate for CHR-ROM data!");
	} else {
		cartridge.CHRfull = malloc(1024 << 10);
	}
	
	if (cartridge.hasTrainer == true) fseek(romfile, 528, SEEK_SET);
		else fseek(romfile, 16, SEEK_SET);
	if (fread(cartridge.PRGfull, 1, (uint32_t)cartridge.PRGcount << 10, romfile) < ((uint32_t)cartridge.PRGcount << 10))
		fatalerr("The ROM appears to be missing PRG data!");
	if (cartridge.CHRcount > 0) if (fread(cartridge.CHRfull, 1, (uint32_t)cartridge.CHRcount << 10, romfile) < (uint32_t)cartridge.CHRcount << 10)
		fatalerr("The ROM appears to be missing CHR data!");	
	fclose(romfile);

	//first init all PRG/CHR bank pointers to the beginning of PRG/CHR memory
	//just in case, to avoid possible crashes from bad pointers
	for (setptr=0; setptr<1024; setptr++) {
		cartridge.CHRbank[setptr] = cartridge.CHRfull;
		cartridge.PRGbank[setptr] = cartridge.PRGfull;
	}

	if (mapperinit(&cartridge) == false) return(false);

	//delete any existing quick save files
	sprintf(&path[0], "%sslot1.mns", &app_path[0]); DeleteFile(&path[0]);
	sprintf(&path[0], "%sslot2.mns", &app_path[0]); DeleteFile(&path[0]);
	sprintf(&path[0], "%sslot3.mns", &app_path[0]); DeleteFile(&path[0]);
	sprintf(&path[0], "%sslot4.mns", &app_path[0]); DeleteFile(&path[0]);

	return(true);
}
